import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np
import cv2
import time
import pickle

opponent = "w"
if opponent != "w" and opponent != "b": raise Exception("Farbe des Gegners falsch festgelegt")

# src beinhaltet die Eckpunkte vom Schachbrett
if True:                                # src laden
    try:
        fileObj = open("src.obj", 'rb')
        src = pickle.load(fileObj)
        fileObj.close()
        print("Eckdaten geladen")
        
    except:
        src = np.float32([ (172,  80),  # links oben
                           (472,  78),  # rechts oben
                           (175,  375), # links unten
                           (470, 382)]) # rechts unten

imSize = 480                   # Bildgröße von Webcam Foto
camera = cv2.VideoCapture(0)

# Square Recognition 
zoom = 5
bwthresh = 18
threshold = 50

# Ist Rochade / en passant möglich?
castling = True
enpassant = 4

out = []  # Liste geschlagener Figuren

# Figuren
B = [["r", "n", "b", "q", "k", "b", "n", "r"], 
     ["p", "p", "p", "p", "p", "p", "p", "p"], 
     ["" , "" , "" , "" , "" , "" , "" , "" ], 
     ["" , "" , "" , "" , "" , "" , "" , "" ], 
     ["" , "" , "" , "" , "" , "" , "" , "" ], 
     ["" , "" , "" , "" , "" , "" , "" , "" ], 
     ["P", "P", "P", "P", "P", "P", "P", "P"], 
     ["R", "N", "B", "Q", "K", "B", "N", "R"]] # print(S[8-y][x-1])

S = B
if opponent == "w": 
    W = B.copy()
    W.reverse()
    for i in range (8):
        W[i].reverse()
    S = W

##### Definitionen Kalibrierung
        
def write (obj, name):          # Wert abspeichern
    fileObj = open(name, 'wb')
    pickle.dump(obj,fileObj)
    fileObj.close()

def read (name):                # Wert auslesen
    fileObj = open(name, 'rb')
    obj = pickle.load(fileObj)
    fileObj.close()
    return obj

def takePicture ():             # Bild aufnehmen
    s, img = camera.read()
    if s: return img
    else: raise Exception("Kamera funktioniert nicht")

def getSrc ():                  # Eckpunkte des Schachbretts grafisch darstellen
    img = takePicture()
    h, w = img.shape[:2]    
        
    dst = np.float32([ (imSize, 0), (0,0), (imSize, imSize), (0, imSize)])  # Zielmatrix 
            
    M = cv2.getPerspectiveTransform(src, dst)                               # Perspektivische
    warped = cv2.warpPerspective(img, M, (w, h), flags=cv2.INTER_LINEAR)    # Korriegierung
    
    # Anzeigen
    f, (ax2) = plt.subplots(1, 1, figsize=(20, 10))
    f.subplots_adjust(hspace=.2, wspace=.05)
    
    ax2.imshow(cv2.flip(warped, 1))
    ax2.set_title('Unwarped Image', fontsize=30)
    
    f, (ax1) = plt.subplots(1, 1, figsize=(20, 10))
    f.subplots_adjust(hspace=.2, wspace=.05)
    ax1.imshow(img)
    
    x = [src[0][0], src[2][0], src[3][0], src[1][0], src[0][0]]
    y = [src[0][1], src[2][1], src[3][1], src[1][1], src[0][1]]
    
    ax1.plot(x, y, color='red', alpha=0.4, linewidth=3, solid_capstyle='round', zorder=2)
    
    ax1.set_ylim([h, 0])
    ax1.set_xlim([0, w])
    ax1.set_title('Ecke: ', fontsize=30)
    
    plt.show()

getSrc()
write(src, "src.obj")

##### Definitionen Bilderkennung

def do (**M):                                     # Bilderkennung durchführen
    global beforeMove
    try:                                                      # Weiß war gerade am Zug, Bilderkennung starten
        if M["mat"] != None:                                           
            afterMove = cropCapture(takePicture())                            # Bild machen und zuschneiden
            
            fields = detectFields(beforeMove, afterMove, castling, enpassant) # Beteiligte Felder ermitteln
            print("Felder: ", fields)
            
            N = writeM(M["mat"], fields)                                      # Änderungen in neue Matrix schreiben
                       
            showAllOnBoard(N)                                                 # Aktuelle Figurenpositionen anzeigen
            
            del beforeMove
            return N

    except KeyError: beforeMove = cropCapture(takePicture(0)) # Schwarz war gerade am Zug, Vorherbild machen
        
def cropCapture (img):                            # Bild zuschneiden
    dst = np.float32([ (imSize, 0), (0,0), (imSize, imSize), (0, imSize)])
    h, w = img.shape[:2]    
    M = cv2.getPerspectiveTransform(src, dst)
    warped  = cv2.warpPerspective(img, M, (w, h), flags=cv2.INTER_LINEAR)
    cropped = warped[int(dst[1][0]):int(dst[2][0]), int(dst[1][1]):int(dst[2][1])]
    flipped = cv2.flip(cropped, 1)
    rotated = cv2.rotate(flipped, cv2.cv2.ROTATE_90_CLOCKWISE) 
    return rotated

def detectFields (im1, im2, castling, enpassant): # Zwei Weißeste Felder finden
    global threshold
    
    diff = cv2.absdiff(im1, im2)                                      # Bilder subrtrahieren
    gray = cv2.cvtColor(diff,cv2.COLOR_BGR2GRAY)                      # Bilder in Schwarz Weiß
    blur = cv2.GaussianBlur(gray, (5,5), 0)                           # Gausscher Unschärfefilter
    _, thresh = cv2.threshold(blur, bwthresh, 255, cv2.THRESH_BINARY) # Helligkeitswerte zu Schwarz oder Weiß auf/- abrunden
    
    fieldSize = int(gray.shape[0] / 8)
    
    if castling:                                   # Rochade                            
        print("Rochade möglich")
        
        # Beteiligte Felder, je nach gespielter Farbe festlegen
        if opponent == "w":
            kingside  = [1,2,3,4]
            queenside = [4,5,6,8]
            
        elif opponent == "b":
            queenside = [1,3,4,5]
            kingside  = [5,6,7,8]
        
        row = 8
        
        # Alle beteiligten Felder einzeln auf durchschnittliche Helligkeit überprüfen
        for x in kingside:
            cropped = thresh[fieldSize*(8-row)+zoom:fieldSize*(9-row)-zoom, fieldSize*(x-1)+zoom:fieldSize*x-zoom] # Bild auf ein einzelnes Feld zuschneiden
            averageCrop = cropped.mean(axis=0).mean(axis=0)   # Durchschnittlichen Farbwert errechnen
             
            if averageCrop < threshold:
                print("Keine Rochade Kingside")
                break
            if x == kingside[len(kingside)-1]:
                print("castling kingside")
                return "castling kingside"
            
        for x in queenside:
            cropped = thresh[fieldSize*(8-row)+zoom:fieldSize*(9-row)-zoom, fieldSize*(x-1)+zoom:fieldSize*x-zoom] # Bild auf ein einzelnes Feld zuschneiden
            averageCrop = cropped.mean(axis=0).mean(axis=0)   # Durchschnittlichen Farbwert errechnen
            
            if averageCrop < threshold:
                print("Keine Rochade Queenside")
                break
            if x == queenside[len(queenside)-1]:
                print("castling queenside")
                return "castling queenside"
            
        print("Keine Rochade!")
    
    # Die Variable en passant gibt an in welcher Spalte das en passant stattfindet (1-8),
    # bei 0 findet kein en passant statt
    if enpassant not in range(0,9): raise Exception ("En passant muss 0 <= x <= 8")
    
    if enpassant !=0 :                             # En passant
        print("En passant möglich")
        
        # Beteiligte Felder
        column = enpassant
        yFields = [4,3]
        
        # Beteiligte Felder in y-Richtung einzeln auf durchschnittliche Helligkeit überprüfen
        for y in yFields:
            cropped = thresh[fieldSize*(8-y)+zoom:fieldSize*(9-y)-zoom, fieldSize*(column-1)+zoom:fieldSize*column-zoom]
            averageCrop = cropped.mean(axis=0).mean(axis=0)
                        
            if averageCrop < threshold:
                print("Kein en passant")
                break
            
            print("Ein Feld für en passant erkannt")
            
            if y == yFields[len(yFields)-1]:
                
                print("Beide Felder erkannt")
                
                # Ermitteln, ob Bauer von rechts, links oder gar nicht schlägt
                try:
                    leftP = thresh[fieldSize*(8-yFields[0])+zoom:fieldSize*(9-yFields[0])-zoom, fieldSize*(column-1-1)+zoom:fieldSize*(column-1)-zoom]
                    left = leftP.mean(axis=0).mean(axis=0)
                    
                    if left > threshold:
                        print("En passent von links, Reihe " + column)
                        return [(column-1, yFields[0])(column, yFields[1]), 1]
                except:
                    print("Kein en passant links")
                    try:
                        rightP = thresh[fieldSize*(8-yFields[0])+zoom:fieldSize*(9-yFields[0])-zoom, fieldSize*(column+1-1)+zoom:fieldSize*(column+1)-zoom]
                        right = rightP.mean(axis=0).mean(axis=0)
                        
                        if right > threshold:
                            print("En passent von rechts, Reihe " + column)
                            return [(column+1, yFields[0])(column, yFields[1]), 1]
                        
                    finally:
                        print("Kein en passant rechts")
                        print("Kein en passant")
                        break
                
    if True:                                       # Nur zwei Felder sind beteiligt
        print("Nur zwei Felder")
    
        field1 = [0, 0]
        field2 = [0, 0]
        
        if gray.mean(axis=0).mean(axis=0) > 0:  # Prüfen, ob es einen Unterschied der zwei verwendeten Bilder gibt
            
        # Höchste Helligkeitswerte, die gefunden werden
            maxAvg1 = 0
            maxAvg2 = 0
            
            for x in range(1,9):
                for y in range(1,9):
                    # Alle Felder nach Helligkeit überprüfen
                    cropped = thresh[fieldSize*(8-y)+zoom:fieldSize*(9-y)-zoom, fieldSize*(x-1)+zoom:fieldSize*x-zoom]
                    averageCrop = cropped.mean(axis=0).mean(axis=0)
                    
                    # Aktuelle Elemente ersetzen den Vorgänger, wenn diese heller sind
                    if averageCrop > threshold:                            
                        if averageCrop > maxAvg1:  
                            field1 = [x, y]
                            maxAvg1 = averageCrop

                        elif averageCrop > maxAvg2:                      
                            field2 = [x, y]
                            maxAvg2 = averageCrop
            
            if field1 == [0, 0] or field2 == [0, 0]:                     # Wenn ein oder zwei Felder leer sind
                if threshold > 1:                     
                    threshold -= 2                                       # Wird der Schwellwert gesenkt
                    return detectFields(im1, im2, castling, enpassant)   # Und die Funktion erneut ausgeführt
                else: raise Exception("Es wurde keine Figur bewegt")     # Es sei denn der Schwellwert ist bereits <= 0, dann gibt es einen Fehler
            else: return (field1, field2)                                # Wenn es für beide Felder einen Wert gibt, werden diese zurück gegeben
        
        else: raise Exception("Bilder sind identisch")

# Figurenbewegung in neue Matrix übertragen
def swap (M, f1, f2):                             # Feldinhalte tauschen   
    f1cont = M[8 - f1[1]] [f1[0] - 1]
    M[8 - f1[1]] [f1[0] - 1] = M[8 - f2[1]] [f2[0] - 1]
    M[8 - f2[1]] [f2[0] - 1] = f1cont
    return M
    
def capture (M, f1, f2):                          # Figur schlagen
    f1Cont = M[8 - f1[1]] [f1[0] - 1]
    f2Cont = M[8 - f2[1]] [f2[0] - 1]
    
    if f1Cont != "" and f2Cont != "":             # Gegnerische Figur wird geschlagen, wenn diese auf beteiligten Felder steht
        if (opponent == "w" and f1Cont.isupper()==False) or (
            opponent == "b" and f1Cont.isupper()):
            # F1 wird geschlagen
            out.append(f1Cont)
            M[8 - f1[1]] [f1[0] - 1] = ""
    
        elif (opponent == "w" and f2Cont.isupper()==False) or (
              opponent == "b" and f2Cont.isupper()):
            # F2 wird geschlagen
            out.append(f2Cont)
            M[8 - f2[1]] [f2[0] - 1] = ""
    return M
    
def writeM (M, felder):                           # Neue Matrix schreiben
    if len(felder) == 2:                             # Normal
        M = capture(M, felder[0], felder[1])
        M = swap(M, felder[0], felder[1])
        
    elif len(felder) == 17:                          # Castling Kingside
        if opponent == "w":
            M = swap(M, [4,8], [2,8])
            M = swap(M, [1,8], [3,8])
            
        if opponent == "b":
            M = swap(M, [5,8], [7,8])
            M = swap(M, [8,8], [6,8])
        
    elif len(felder) == 18:                          # Castling Queenside
        if opponent == "w":
            M = swap(M, [4,8], [6,8])
            M = swap(M, [8,8], [5,8])
    
        if opponent == "b":
            M = swap(M, [5,8], [3,8])
            M = swap(M, [1,8], [4,8])
        
    elif len(felder) == 3:                           # En passant
        M = capture(M, felder[0], [felder[1][0], 4]) # Schlage Figur auf Feld (x(Feld2), 4)
        M = swap(M, felder[0], felder[1])
        
    else: print("Falsche Felder")
    return M

# Alle Figuren anzeigen
def showAllOnBoard (M):                           
    fig, axarr = plt.subplots(8,8, figsize=(50,50))
      
    for x in range(1,9):
        for y in range(1,9):
            cont = M[8-y][x-1]                     # Matrix der aktuellen Figurenpositionen
            if cont != "":
                if cont.isupper(): axarr[8-y][x-1].imshow(mpimg.imread("pieces/"+ cont   +".png")) # Weiße Figuren 
                else:              axarr[8-y][x-1].imshow(mpimg.imread("pieces/"+ cont*2 +".png")) # Schwarze Figuren
   
    [axi.set_axis_off() for axi in axarr.ravel()]
    fig.tight_layout(h_pad=0)
    plt.show()

##### Programm ausführen

do()
for x in range(10):
    print("white to move")
    time.sleep(2)
    do(mat=S)
    
    print("black to move")
    time.sleep(2)
    do()